import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_min():
    example = np.asarray(
        [
            [
                [0, 0, 0, 1],
                [0, 0, 3, 1],
            ],
            [[0, 0, 3, 1], [1, 1, 1, 1]],
        ]
    )

    gpu_example = cle.push(example)

    assert gpu_example.min() == example.min()


def test_min_max_xyz():
    example = np.asarray(
        [
            [
                [0, 0, 0, 1],
                [0, 0, 3, 1],
            ],
            [[0, 0, 3, 1], [1, 1, 1, 1]],
        ]
    )

    gpu_example = cle.push(example)

    print(gpu_example.min(axis=0))
    print(example.min(axis=0))
    print("---")
    print(gpu_example.min(axis=1))
    print(example.min(axis=1))
    print("---")
    print(gpu_example.min(axis=2))
    print(example.min(axis=2))
    print("---")
    print(gpu_example.max(axis=0))
    print(example.max(axis=0))
    print("---")
    print(gpu_example.max(axis=1))
    print(example.max(axis=1))
    print("---")
    print(gpu_example.max(axis=2))
    print(example.max(axis=2))

    assert np.allclose(gpu_example.min(axis=0), example.min(axis=0))
    assert np.allclose(gpu_example.min(axis=1), example.min(axis=1))
    assert np.allclose(gpu_example.min(axis=2), example.min(axis=2))
    assert np.allclose(gpu_example.min(), example.min())

    assert np.allclose(gpu_example.max(axis=0), example.max(axis=0))
    assert np.allclose(gpu_example.max(axis=1), example.max(axis=1))
    assert np.allclose(gpu_example.max(axis=2), example.max(axis=2))
    assert np.allclose(gpu_example.max(), example.max())

    assert np.allclose(gpu_example.sum(axis=0), example.sum(axis=0))
    assert np.allclose(gpu_example.sum(axis=1), example.sum(axis=1))
    assert np.allclose(gpu_example.sum(axis=2), example.sum(axis=2))
    assert np.allclose(gpu_example.sum(), example.sum())


def test_min_out():
    example = np.asarray(
        [
            [
                [0, 0, 0, 1],
                [0, 0, 3, 1],
            ],
            [[0, 0, 3, 1], [1, 1, 1, 1]],
        ]
    )

    gpu_example = cle.push(example)

    minimum = example.min(axis=0)
    copy = minimum.copy()
    gpu_example.min(axis=0, out=copy)
    assert np.allclose(minimum, copy)

    minimum = example.min(axis=1)
    copy = minimum.copy()
    gpu_example.min(axis=1, out=copy)
    assert np.allclose(minimum, copy)

    minimum = example.min(axis=2)
    copy = minimum.copy()
    gpu_example.min(axis=2, out=copy)
    assert np.allclose(minimum, copy)


def test_max_out():
    example = np.asarray(
        [
            [
                [0, 0, 0, 1],
                [0, 0, 3, 1],
            ],
            [[0, 0, 3, 1], [1, 1, 1, 1]],
        ]
    )

    gpu_example = cle.push(example)

    maximum = example.max(axis=0)
    copy = maximum.copy()
    gpu_example.max(axis=0, out=copy)
    assert np.allclose(maximum, copy)

    maximum = example.max(axis=1)
    copy = maximum.copy()
    gpu_example.max(axis=1, out=copy)
    assert np.allclose(maximum, copy)

    maximum = example.max(axis=2)
    copy = maximum.copy()
    gpu_example.max(axis=2, out=copy)
    assert np.allclose(maximum, copy)


def test_sum_out():
    example = np.asarray(
        [
            [
                [0, 0, 0, 1],
                [0, 0, 3, 1],
            ],
            [[0, 0, 3, 1], [1, 1, 1, 1]],
        ]
    )

    gpu_example = cle.push(example)

    sum = example.sum(axis=0)
    copy = sum.copy()
    gpu_example.sum(axis=0, out=copy)
    assert np.allclose(sum, copy)

    sum = example.sum(axis=1)
    copy = sum.copy()
    gpu_example.sum(axis=1, out=copy)
    assert np.allclose(sum, copy)

    sum = example.sum(axis=2)
    copy = sum.copy()
    gpu_example.sum(axis=2, out=copy)
    assert np.allclose(sum, copy)



#
# This code is auto-generated from CLIc 'cle::tier7.hpp' file, do not edit manually.
#

import importlib
import warnings
from typing import Optional

import numpy as np

from ._array import Image
from ._core import Device
from ._decorators import plugin_function

clic = importlib.import_module('._pyclesperanto', package='pyclesperanto')

@plugin_function
def affine_transform(
    input_image: Image,
    output_image: Optional[Image] =None,
    transform_matrix: Optional[list] =None,
    interpolate: bool =False,
    resize: bool =False,
    device: Optional[Device] =None
) -> Image:
    """Apply an affine transformation matrix to an array and return the result.  The
    transformation matrix must be 3x3 or 4x4 stored as a 1D array.  The matrix
    should be row-major, i.e. the first 3 elements are the first row of the matrix.
    If no matrix is given, the identity matrix will be used.

    Parameters
    ----------
    input_image: Image 
        Input Array to be transformed.
    output_image: Optional[Image] (= None)
        Output Array.
    transform_matrix: Optional[list] (= None)
        Affine transformation matrix (3x3 or 4x4).
    interpolate: bool (= False)
        If true, bi/trilinear interpolation will be applied, if hardware allows.
    resize: bool (= False)
        Automatically determines the size of the output depending on the rotation angles.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._affine_transform(device, input_image, output_image, transform_matrix, interpolate, resize)

@plugin_function(categories=["label", "in assistant"])
def eroded_otsu_labeling(
    input_image: Image,
    output_image: Optional[Image] =None,
    number_of_erosions: int =5,
    outline_sigma: float =2,
    device: Optional[Device] =None
) -> Image:
    """Segments and labels an image using blurring, Otsu-thresholding, binary erosion
    and  masked Voronoi-labeling.  After bluring and Otsu-thresholding the image,
    iterative binary erosion is applied.  Objects in the eroded image are labeled
    and the labels are extended to fit again into  the initial binary image using
    masked-Voronoi labeling.  This function is similar to voronoi_otsu_labeling. It
    is intended to deal better in  case labels of objects swapping into each other
    if objects are dense. Like when using  Voronoi-Otsu-labeling, small objects may
    disappear when applying this operation.  This function is inspired by a similar
    implementation in Java by Jan Brocher (Biovoxxel) [0] [1]

    Parameters
    ----------
    input_image: Image 
        Input Array to be transformed.
    output_image: Optional[Image] (= None)
        Output Array.
    number_of_erosions: int (= 5)
        Number of iteration of erosion.
    outline_sigma: float (= 2)
        Gaussian blur sigma applied before Otsu thresholding.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://github.com/biovoxxel/bv3dbox (BV_LabelSplitter.java#L83)
    [2] https://zenodo.org/badge/latestdoi/434949702
    """
    return clic._eroded_otsu_labeling(device, input_image, output_image, int(number_of_erosions), float(outline_sigma))

@plugin_function(categories=["transform", "in assistant"])
def rigid_transform(
    input_image: Image,
    output_image: Optional[Image] =None,
    translate_x: float =0,
    translate_y: float =0,
    translate_z: float =0,
    angle_x: float =0,
    angle_y: float =0,
    angle_z: float =0,
    centered: bool =True,
    interpolate: bool =False,
    resize: bool =False,
    device: Optional[Device] =None
) -> Image:
    """Translate the image by a given vector and rotate it by given angles. Angles are
    given in degrees. To convert radians to degrees, use this formula:
    angle_in_degrees = angle_in_radians / numpy.pi * 180.0

    Parameters
    ----------
    input_image: Image 
        Input Array to be transformed.
    output_image: Optional[Image] (= None)
        Output Array.
    translate_x: float (= 0)
        Translation along x axis in pixels.
    translate_y: float (= 0)
        Translation along y axis in pixels.
    translate_z: float (= 0)
        Translation along z axis in pixels.
    angle_x: float (= 0)
        Rotation around x axis in radians.
    angle_y: float (= 0)
        Rotation around y axis in radians.
    angle_z: float (= 0)
        Rotation around z axis in radians.
    centered: bool (= True)
        If true, rotate image around center, else around the origin.
    interpolate: bool (= False)
        If true, bi/trilinear interpolation will be applied, if hardware allows.
    resize: bool (= False)
        Automatically determines the size of the output depending on the rotation angles.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._rigid_transform(device, input_image, output_image, float(translate_x), float(translate_y), float(translate_z), float(angle_x), float(angle_y), float(angle_z), centered, interpolate, resize)

@plugin_function(categories=["transform", "in assistant"])
def rotate(
    input_image: Image,
    output_image: Optional[Image] =None,
    angle_x: float =0,
    angle_y: float =0,
    angle_z: float =0,
    centered: bool =True,
    interpolate: bool =False,
    resize: bool =False,
    device: Optional[Device] =None
) -> Image:
    """Rotate the image by given angles. Angles are given in degrees. To convert
    radians to degrees, use this formula: angle_in_degrees = angle_in_radians /
    numpy.pi * 180.0

    Parameters
    ----------
    input_image: Image 
        Input Array to be rotated.
    output_image: Optional[Image] (= None)
        Output Array.
    angle_x: float (= 0)
        Rotation around x axis in degrees.
    angle_y: float (= 0)
        Rotation around y axis in degrees.
    angle_z: float (= 0)
        Rotation around z axis in degrees.
    centered: bool (= True)
        If true, rotate image around center, else around the origin.
    interpolate: bool (= False)
        If true, bi/trilinear interpolation will be applied, if hardware allows.
    resize: bool (= False)
        Automatically determines the size of the output depending on the rotation angles.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._rotate(device, input_image, output_image, float(angle_x), float(angle_y), float(angle_z), centered, interpolate, resize)

@plugin_function(categories=["transform", "in assistant"])
def scale(
    input_image: Image,
    output_image: Optional[Image] =None,
    factor_x: float =1,
    factor_y: float =1,
    factor_z: float =1,
    centered: bool =True,
    interpolate: bool =False,
    resize: bool =False,
    device: Optional[Device] =None
) -> Image:
    """Scale the image by given factors.

    Parameters
    ----------
    input_image: Image 
        Input Array to be scaleded.
    output_image: Optional[Image] (= None)
        Output Array.
    factor_x: float (= 1)
        Scaling along x axis.
    factor_y: float (= 1)
        Scaling along y axis.
    factor_z: float (= 1)
        Scaling along z axis.
    centered: bool (= True)
        If true, the image will be scaled to the center of the image.
    interpolate: bool (= False)
        If true, bi/trilinear interplation will be applied.
    resize: bool (= False)
        Automatically determines output size image.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._scale(device, input_image, output_image, float(factor_x), float(factor_y), float(factor_z), centered, interpolate, resize)

@plugin_function(categories=["transform", "in assistant"])
def translate(
    input_image: Image,
    output_image: Optional[Image] =None,
    translate_x: float =0,
    translate_y: float =0,
    translate_z: float =0,
    interpolate: bool =False,
    device: Optional[Device] =None
) -> Image:
    """Translate the image by a given vector.

    Parameters
    ----------
    input_image: Image 
        Input Array to be translated.
    output_image: Optional[Image] (= None)
        Output Array.
    translate_x: float (= 0)
        Translation along x axis in pixels.
    translate_y: float (= 0)
        Translation along y axis in pixels.
    translate_z: float (= 0)
        Translation along z axis in pixels.
    interpolate: bool (= False)
        If true, bi/trilinear interplation will be applied.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._translate(device, input_image, output_image, float(translate_x), float(translate_y), float(translate_z), interpolate)

@plugin_function(categories=["label processing", "in assistant"])
def closing_labels(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius: int =0,
    device: Optional[Device] =None
) -> Image:
    """Apply a morphological closing operation to a label image. The operation consists
    of iterative dilation and erosion of the labels. With every iteration, box and
    diamond/sphere structuring elements are used and thus, the operation has an
    octagon as structuring element. Notes * This operation assumes input images are
    isotropic.

    Parameters
    ----------
    input_image: Image 
        Input label Array.
    output_image: Optional[Image] (= None)
        Output label Array.
    radius: int (= 0)
        Radius size for the closing.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._closing_labels(device, input_image, output_image, int(radius))

@plugin_function(categories=["label processing", "in assistant"])
def erode_connected_labels(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius: int =1,
    device: Optional[Device] =None
) -> Image:
    """Erodes labels to a smaller size. Note: Depending on the label image and the
    radius,  labels may disappear and labels may split into multiple islands. Thus,
    overlapping labels of input and output may  not have the same identifier.

    Parameters
    ----------
    input_image: Image 
        result
    output_image: Optional[Image] (= None)
        
    radius: int (= 1)
        
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._erode_connected_labels(device, input_image, output_image, int(radius))

@plugin_function(categories=["label processing", "in assistant"])
def opening_labels(
    input_image: Image,
    output_image: Optional[Image] =None,
    radius: int =0,
    device: Optional[Device] =None
) -> Image:
    """Apply a morphological opening operation to a label image. The operation consists
    of iterative erosion and dilation of the labels. With every iteration, box and
    diamond/sphere structuring elements are used and thus, the operation has an
    octagon as structuring element. Notes * This operation assumes input images are
    isotropic.

    Parameters
    ----------
    input_image: Image 
        Input label Array.
    output_image: Optional[Image] (= None)
        Output label Array.
    radius: int (= 0)
        Radius size for the opening.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image
    """
    return clic._opening_labels(device, input_image, output_image, int(radius))

@plugin_function(categories=["label", "in assistant", "bia-bob-suggestion"])
def voronoi_otsu_labeling(
    input_image: Image,
    output_image: Optional[Image] =None,
    spot_sigma: float =2,
    outline_sigma: float =2,
    device: Optional[Device] =None
) -> Image:
    """Labels objects directly from greyvalue images. The two sigma parameters allow
    tuning the segmentation result. Under the hood, this filter applies two Gaussian
    blurs, spot detection, Otsuthresholding [2] and Voronoilabeling [3]. The
    thresholded binary image is flooded using the Voronoi tesselation approach
    starting from the found local maxima. Notes * This operation assumes input
    images are isotropic.

    Parameters
    ----------
    input_image: Image 
        Input intensity Array.
    output_image: Optional[Image] (= None)
        Output label Array.
    spot_sigma: float (= 2)
        Controls how close detected cells can be.
    outline_sigma: float (= 2)
        Controls how precise segmented objects are outlined.
    device: Optional[Device] (= None)
        Device to perform the operation on.

    Returns
    -------
    Image

    References
    ----------
    [1] https://clij.github.io/clij2-docs/reference_voronoiOtsuLabeling
    [2] https://ieeexplore.ieee.org/document/4310076
    [3] https://en.wikipedia.org/wiki/Voronoi_diagram
    """
    return clic._voronoi_otsu_labeling(device, input_image, output_image, float(spot_sigma), float(outline_sigma))


import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_binary_and_1():
    test = cle.push(np.asarray([[1, 0], [1, 0]]))

    test1 = cle.push(np.asarray([[1, 1], [0, 0]]))

    test2 = cle.create(test)
    cle.binary_and(test, test1, test2)

    print(test2)

    a = cle.pull(test2)
    assert np.min(a) == 0
    assert np.max(a) == 1
    assert np.mean(a) == 0.25


def test_binary_and_2():
    a = np.asarray([[1, 0], [1, 0]])
    b = np.asarray([[1, 1], [0, 0]])
    gpu_a = cle.push(a)
    gpu_b = cle.push(b)
    gpu_c = cle.create(gpu_a)
    cle.binary_and(gpu_a, gpu_b, gpu_c)

    result = cle.pull(gpu_c)
    assert np.allclose(result, a & b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_nonzero_maximum_box():
    test = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 1, 1, 2, 0],
                [0, 2, 2, 3, 0],
                [0, 3, 3, 4, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    reference = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 2, 3, 3, 0],
                [0, 3, 4, 4, 0],
                [0, 3, 4, 4, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    result = cle.create(test)
    flag = cle.create((1, 1, 1))

    # as nonzero filters don't touch zero values, we need to initialize the result in advance
    cle.set(result, 0)

    cle.nonzero_maximum(test, flag, result)

    print(result)

    a = cle.pull(result)
    b = cle.pull(reference)
    assert np.allclose(a, b, atol=0.00001)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_minimum_box():
    test1 = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 1, 1, 1, 0],
                [0, 1, 1, 1, 0],
                [0, 1, 1, 1, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    reference = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 1, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    result = cle.create(test1)
    cle.minimum(test1, result, 1, 1, 0)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_clip_min_max():
    test = [[0, 1], [2, 3]]
    reference = [[1, 1], [2, 2]]

    result = cle.clip(test, a_min=1, a_max=2)

    assert np.array_equal(result, reference)


def test_clip_max():
    test = [[0, 1], [2, 3]]
    reference = [[0, 1], [2, 2]]

    result = cle.clip(test, a_min=0, a_max=2)

    assert np.array_equal(result, reference)


def test_clip_min():
    test = [[0, 1], [2, 3]]
    reference = [[1, 1], [2, 3]]

    result = cle.clip(test, a_min=1, a_max=3)

    assert np.array_equal(result, reference)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_setitem_3d():
    data = np.asarray([[[1.0, 2.0, 3.0], [1.0, 2.0, 3.0]]]).astype(np.float32)
    ref = np.asarray([[[4.0, 2.0, 3.0], [4.0, 4.0, 3.0]]]).astype(np.float32)

    cl_data = cle.push(data)

    positions = (np.asarray([0, 0, 0]), np.asarray([0, 1, 1]), np.asarray([0, 1, 0]))

    cl_data[positions] = 4

    assert np.allclose(cl_data, ref)


def test_setitem_2d():
    data = np.asarray([[1.0, 2.0, 3.0], [1.0, 2.0, 3.0]]).astype(np.float32)
    ref = np.asarray([[4.0, 2.0, 3.0], [4.0, 4.0, 3.0]]).astype(np.float32)

    cl_data = cle.push(data)

    positions = (np.asarray([0, 1, 1]), np.asarray([0, 1, 0]))

    cl_data[positions] = 4

    assert np.allclose(cl_data, ref)



from unittest.mock import patch

import matplotlib.pyplot as plt
import numpy as np

import pyclesperanto as cle


@patch("matplotlib.pyplot.show")
def test_imshow(mock_show):
    image = cle.push(np.random.random((512, 512)))
    cle.imshow(image)


@patch("matplotlib.pyplot.show")
def test_imshow_3d(mock_show):
    image = cle.push(np.random.random((10, 512, 512)))
    cle.imshow(image, colorbar=True)
    plt.show()
    plt.close()


@patch("matplotlib.pyplot.show")
def test_imshow_label(mock_show):
    image = cle.push(np.random.random((512, 512)))
    label = cle.connected_components_labeling(image > 0.8)
    cle.imshow(label, labels=True)
    plt.show()



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_gamma_correction():
    test = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 50, 0, 5, 0],
                [0, 0, 100, 0, 0],
                [0, 30, 0, 10, 0],
                [0, 0, 0, 0, 0],
            ]
        ).astype(float)
    )

    result = cle.create(test, dtype=float)
    cle.gamma_correction(test, result, 0.5)

    a = cle.pull(result)

    print(a)
    print(np.mean(a))

    print(np.abs(np.min(a)))
    print(np.abs(np.max(a) - 100))
    print(np.abs(np.mean(a) - 11.1786))

    assert np.abs(np.min(a)) < 0.001
    assert np.abs(np.max(a) - 100) < 0.001
    assert np.abs(np.mean(a) - 11.1786) < 0.001



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")

source = np.asarray([0, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0])
reference = np.asarray([0, 1, 0, 2, 0, 0, 3, 4, 0, 0, 5, 0])


def block_enum(source, blocksize):
    flagged_indices = cle.push(source)
    max_label = source.shape[0] - 1

    block_sums = cle.create([int((int(max_label) + 1) / blocksize) + 1])
    cle.sum_reduction_x(flagged_indices, block_sums, blocksize)

    # distribute new numbers
    new_indices = cle.create([int(max_label) + 1])
    cle.block_enumerate(flagged_indices, block_sums, new_indices, blocksize)

    return cle.pull(new_indices)


def test_block_enumerate():
    result = block_enum(source, 4)
    print(result)
    print(reference)
    assert np.array_equal(result, reference)

    result = block_enum(source, 2)
    print(result)
    print(reference)
    assert np.array_equal(result, reference)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_add_image_and_scalar():
    data = np.arange(100).reshape(10, 10)
    # push an array to the GPU
    flip = cle.push(data)
    assert flip.shape == (10, 10)
    assert isinstance(flip, cle.Array)
    # create memory for the output
    flop = cle.create_like(data)
    assert flop.shape == (10, 10)
    assert isinstance(flop, cle.Array)
    # add a constant to all pixels
    cle.add_image_and_scalar(flip, flop, 100.0)
    # Note the transposition!
    np.testing.assert_allclose(data + 100, flop.get())



import warnings
from typing import Optional, Union

import numpy as np

from ._pyclesperanto import _BackendManager as BackendManager
from ._pyclesperanto import _Device as Device


class _current_device:
    _instance: Optional[Device] = None


def get_device() -> Device:
    """Return the current device instance

    Returns
    -------
    device : Device
    """
    return _current_device._instance or select_device()


def select_device(device_id: Union[str, int] = "", device_type: str = "all") -> Device:
    """Select a device by 'name' or 'index' and by 'type', and store it as the current device

    If selecting the device by string, the function compares the device name and substring.
    (e.g. "NVIDIA", "RTX", "Iris", etc. will match the device name "NVIDIA RTX 2080" or "Intel Iris Pro")
    If selecting the device by index, the function will select the device at the given index in the list
    of available devices. (e.g. 0, 1, 2, etc. will select the first, second, third, etc. device in the list)
    If device_id is an empty string, the function will select the first available device.
    The device_type enables selecting the type of device to be selected (e.g. "all", "cpu", "gpu")
    To retrieve a list of available devices, use `list_available_devices()`

    Parameters
    ----------
    device_id : Union[str, int], default = ""
        Substring of device name or device index.
    device_type : str, default = "all"
        Type of device to be selected (e.g. "all", "cpu", "gpu")

    Returns
    -------
    device : Device
    """
    if isinstance(device_id, str):
        device = BackendManager.get_backend().getDeviceFromName(device_id, device_type)
    elif isinstance(device_id, int):
        device = BackendManager.get_backend().getDeviceFromIndex(device_id, device_type)
    else:
        raise ValueError(
            f"'{device_id}' is not a supported device_id. Please use either a string or an integer."
        )
    if _current_device._instance and device == _current_device._instance:
        return _current_device._instance
    _current_device._instance = device
    return device


def list_available_devices(device_type: str = "all") -> list:
    """Retrieve a list of names of available devices

    Will search the system for backend compatible device available and return a list of their names.
    This will NOT set the device!
    Use 'select_device' to select devices.
    Use 'get_device' to retrieve the current device.

    Parameters
    ----------
    device_type : str, default = "all"
        Type of device to be selected (e.g. "all", "cpu", "gpu")

    Returns
    -------
    name list : list[str]
    """
    dev_list = list(BackendManager.get_backend().getDevicesList(type=device_type))
    if not dev_list:
        warnings.warn(
            "No device available. Please install either OpenCL or CUDA on your system.",
            RuntimeWarning,
        )
    return dev_list


def list_available_backends() -> list:
    """Retrieve a list of names of available backends

    Will test system for available backends installed and return a list of their names.

    Returns
    -------
    name list : list[str]
    """
    back_list = list(BackendManager.get_backends_list())
    if not back_list:
        warnings.warn(
            "No backend available. Please install either OpenCL or CUDA on your system.",
            RuntimeWarning,
        )
    return back_list


def select_backend(backend: str = "opencl") -> str:
    """select backend

    Select the backend used by pyclesperanto, OpenCL or CUDA.
    Default is OpenCL.

    Parameters
    ----------
    type : str, default = "opencl"
        determine the backend to use between opencl and cuda
    """

    # enforce lowercase for backend_type
    backend = backend.lower()
    # is backend_type is different than "cuda" or "opencl", raise an error
    if backend not in ["cuda", "opencl"]:
        raise ValueError(
            f"'{backend}' is not a supported Backend. Please use either 'opencl' or 'cuda'."
        )
    BackendManager.set_backend(backend=backend)
    # reset current device to default one
    select_device()
    return f"{BackendManager.get_backend()} selected."


def wait_for_kernel_to_finish(flag: bool = True, device: Device = None):
    """Wait for kernel to finish

    Enforce the system to wait for the kernel to finish before continuing. Introducing a
    slowdown in the workflow. This is useful for debugging purposes, benchmarking and
    profiling, as well as for complex workflows where the order of operations is important.

    Parameters
    ----------
    flag : bool, default = True
        if True, wait for kernel to finish
    device : Device, default = None
        the device to set the flag. If None, set it to the current device
    """
    if device is None:
        get_device().set_wait_to_finish(flag)
    else:
        device.set_wait_to_finish(flag)


def default_initialisation():
    """Set default backend and device"""
    backends = list_available_backends()
    if backends:
        _ = select_backend(backends[-1])
    else:
        warnings.warn(
            "No GPU backend found.\n\n"
            "pyclesperanto requires either CUDA or OpenCL libraries to be installed on your system to work.\n"
            "Please ensure you have the appropriate drivers installed and up-to-date.\n\n"
            "Alternatively, you may need to install the following additional package:\n"
            "- MacOS: `conda install -c conda-forge ocl_icd_wrapper_apple`\n"
            "- Linux: `conda install -c conda-forge ocl-icd-system`",
            RuntimeWarning,
        )


def info():
    """Print information about the devices available on the system"""
    device_info = [
        f"{idx} - {select_device(device).info}"
        for idx, device in enumerate(list_available_devices())
    ]
    print("".join(device_info))



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_local_cross_correlation():
    test = cle.push(np.asarray([[0, 0, 0], [0, 1, 0], [0, 0, 0]]))

    test1 = cle.push(np.asarray([[0, 1, 0], [1, 2, 1], [0, 1, 0]]))

    reference = cle.push(
        np.asarray(
            [
                [0.0, 0.35355344, 0.0],
                [0.35355344, 0.7071069, 0.35355344],
                [0.0, 0.35355344, 0.0],
            ]
        )
    )

    test2 = cle.create(test, dtype=float)
    cle.local_cross_correlation(test, test1, test2)

    print(test2)

    a = cle.pull(test2)
    b = cle.pull(reference)
    assert np.allclose(a, b, atol=0.001)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_absolute():
    test = cle.push(np.asarray([[1, -1], [1, -1]]))

    test2 = cle.create(test)
    cle.absolute(test, test2)

    print(test2)

    a = cle.pull(test2)
    assert np.min(a) == 1
    assert np.max(a) == 1
    assert np.mean(a) == 1


def test_absolute1():
    gpu_a = cle.push(np.asarray([[1, -1], [1, -1]]))
    gpu_b = cle.create(gpu_a)
    cle.absolute(gpu_a, gpu_b)
    assert np.all(cle.pull(gpu_b) == 1)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_median_sphere():
    test1 = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [9, 9, 9, 0, 0],
                [9, 9, 9, 0, 0],
                [9, 9, 9, 0, 0],
            ]
        )
    )

    reference = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [9, 9, 9, 0, 0],
                [9, 9, 9, 0, 0],
                [9, 9, 9, 0, 0],
            ]
        )
    )

    result = cle.create(test1)
    cle.median(test1, result, 1, 1, 0, "sphere")

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.allclose(a, b, 0.0001)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_voronoi_labeling():
    gpu_input = cle.push(
        np.asarray(
            [
                [
                    [0, 0, 0, 0, 0, 0],
                    [0, 1, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0],
                    [0, 1, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0, 0],
                ]
            ]
        )
    )

    gpu_reference = cle.push(
        np.asarray(
            [
                [
                    [1, 1, 1, 2, 2, 2],
                    [1, 1, 1, 2, 2, 2],
                    [1, 1, 1, 2, 2, 2],
                    [3, 3, 3, 4, 4, 4],
                    [3, 3, 3, 4, 4, 4],
                    [3, 3, 3, 4, 4, 4],
                ]
            ]
        )
    )

    gpu_output = cle.voronoi_labeling(gpu_input)

    a = cle.pull(gpu_output)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)



import numpy as np
import pytest

import pyclesperanto as cle

cle.select_device("TX")


# @pytest.mark.skip(reason="Fails on github CI but passes locally")
def test_sign():
    data = np.asarray([-np.inf, np.inf, 0, 1, -1, np.nan])

    print(np.sign(data))
    print(cle.sign(data))

    # we exclude nan from the test, because numpy cannot compare it
    assert np.allclose(np.sign(data)[0:4], cle.sign(data)[0:4])



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_difference_of_gaussian():
    test = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 100, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    result = cle.create(test, dtype=float)
    cle.difference_of_gaussian(test, result, 1, 1, 0, 2, 2, 0)

    print(result)

    a = cle.pull(result)
    assert np.min(a) < -1.15
    assert np.min(a) > -1.18
    assert np.max(a) > 11.9
    assert np.max(a) < 12



import numpy as np
from skimage.io import imread

import pyclesperanto as cle

cle.select_device("TX")


def test_exclude_labels_on_edges_2d():
    gpu_input = cle.push(
        np.asarray(
            [
                [0, 0, 2, 0, 0, 0, 0],
                [0, 1, 2, 0, 7, 0, 0],
                [0, 1, 0, 0, 7, 5, 5],
                [8, 8, 8, 0, 0, 0, 0],
                [0, 4, 4, 0, 3, 0, 0],
                [0, 4, 4, 6, 0, 0, 0],
            ]
        )
    )

    gpu_reference = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 3, 0, 0],
                [0, 1, 0, 0, 3, 0, 0],
                [0, 0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 2, 0, 0],
                [0, 0, 0, 0, 0, 0, 0],
            ]
        )
    )

    gpu_output = cle.exclude_labels_on_edges(gpu_input)

    a = cle.pull(gpu_output)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)


def test_exclude_labels_on_edges_3d():
    gpu_input = cle.push(
        np.asarray(
            [
                [
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 9, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                ],
                [
                    [0, 0, 2, 0, 0, 0, 0],
                    [0, 1, 2, 0, 7, 0, 0],
                    [0, 1, 0, 0, 7, 5, 5],
                    [8, 8, 8, 0, 0, 0, 0],
                    [0, 4, 4, 9, 3, 0, 0],
                    [0, 4, 4, 6, 0, 0, 0],
                ],
                [
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 7, 0, 0],
                    [0, 0, 0, 0, 7, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                ],
            ]
        )
    )

    gpu_reference = cle.push(
        np.asarray(
            [
                [
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                ],
                [
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 1, 0, 0, 0, 0, 0],
                    [0, 1, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 2, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                ],
                [
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                ],
            ]
        )
    )

    gpu_output = cle.exclude_labels_on_edges(gpu_input)

    a = cle.pull(gpu_output)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)


def test_exclude_labels_on_edges_blobs():
    # initialize GPU
    cle.select_device("TX")
    print("Used GPU: " + cle.get_device().name)

    # load data
    filename = "https://samples.fiji.sc/blobs.png"
    image = np.squeeze(imread(filename))
    print("Loaded image size: " + str(image.shape))

    # push image to GPU memory
    input = cle.push(image)
    print("Image size in GPU: " + str(input.shape))

    # process the image
    blurred = cle.gaussian_blur(image, sigma_x=1, sigma_y=1)
    binary = cle.threshold_otsu(blurred)
    labeled = cle.connected_components_labeling(binary)

    wo_edges = cle.exclude_labels_on_edges(labeled)

    # The maxmium intensity in a label image corresponds to the number of objects
    num_labels = cle.maximum_of_all_pixels(wo_edges)

    # print out result
    print("Num objects in the image: " + str(num_labels))

    assert num_labels == 44


def test_exclude_labels_on_edges_blobs_2():
    # initialize GPU
    cle.select_device("GTX")
    print("Used GPU: " + cle.get_device().name)

    # load data
    filename = "https://samples.fiji.sc/blobs.png"
    image = np.squeeze(imread(filename))
    print("Loaded image size: " + str(image.shape))

    # push image to GPU memory
    input = cle.push(image)
    print("Image size in GPU: " + str(input.shape))

    # process the image
    blurred = cle.gaussian_blur(image, sigma_x=1, sigma_y=1)
    binary = cle.threshold_otsu(blurred)
    labeled = cle.connected_components_labeling(binary)

    wo_edges = cle.exclude_labels_on_edges(labeled, exclude_y=False)

    # The maxmium intensity in a label image corresponds to the number of objects
    num_labels = cle.maximum_of_all_pixels(wo_edges)

    # print out result
    print("Num objects in the image: " + str(num_labels))

    assert num_labels == 52

    wo_edges = cle.exclude_labels_on_edges(labeled, exclude_x=False)

    # The maxmium intensity in a label image corresponds to the number of objects
    num_labels = cle.maximum_of_all_pixels(wo_edges)

    # print out result
    print("Num objects in the image: " + str(num_labels))

    assert num_labels == 53



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_histogram():
    test = cle.push(np.asarray([[1, 2, 4, 4, 2, 3], [3, 3, 4, 4, 5, 5]]))

    ref_histogram = [1, 2, 3, 4, 2]

    my_histogram = cle.histogram(test, nbins=5, min=1, max=5)

    print(my_histogram)

    a = cle.pull(my_histogram)
    assert np.allclose(a, ref_histogram)


def test_histogram_3d():
    test = cle.push(np.asarray([[1, 2, 4, 4, 2, 3], [3, 3, 4, 4, 5, 5]]))

    ref_histogram = [1, 2, 3, 4, 2]

    my_histogram = cle.histogram(test, nbins=5, min=1, max=5)

    print(my_histogram)

    a = cle.pull(my_histogram)
    assert np.allclose(a, ref_histogram)


def test_histogram_3d_2():
    test = cle.push(np.asarray([[[1, 2, 4], [4, 2, 3]], [[3, 3, 4], [4, 5, 5]]]))

    ref_histogram = [1, 2, 3, 4, 2]

    my_histogram = cle.histogram(test, nbins=5, min=1, max=5)

    print(my_histogram)

    a = cle.pull(my_histogram)
    assert np.allclose(a, ref_histogram)


# def test_histogram_against_scikit_image():
#     from skimage.data import camera
#     image = camera()
#     from skimage import exposure
#     hist, bc = exposure.histogram(image.ravel(), 256, source_range="image")
#     print(hist)
#     gpu_image = cle.push(image.astype(int))
#     gpu_hist = cle.histogram(gpu_image, nbins=256, min=0, max=gpu_image.max())
#     print(cle.pull(gpu_hist))
#     assert np.allclose(hist, cle.pull(gpu_hist))



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_minimum_images():
    test1 = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 1, 2, 3, 0],
                [0, 3, 3, 4, 0],
                [0, 4, 4, 5, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )
    test2 = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 2, 1, 3, 0],
                [0, 2, 1, 3, 0],
                [0, 1, 1, 3, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    reference = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 1, 1, 3, 0],
                [0, 2, 1, 3, 0],
                [0, 1, 1, 3, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    result = cle.create(test1)
    cle.minimum_images(test1, test2, result)

    a = cle.pull(result)
    b = cle.pull(reference)
    print(a)

    assert np.array_equal(a, b)


def test_minimum_images_with_types():
    test1 = cle.push(
        np.asarray(
            [
                [0, 3, 3, 4, 0],
            ]
        )
    )
    test2 = np.asarray(
        [
            [0, 2, 1, 3, 0],
        ]
    )

    reference = cle.push(
        np.asarray(
            [
                [0, 2, 1, 3, 0],
            ]
        )
    )

    result1 = cle.minimum_images(test1.astype(int), test2.astype(int))
    result2 = cle.minimum_images(test1.astype(float), test2.astype(float))
    result3 = cle.minimum_images(test1.astype(int), test2.astype(float))
    result4 = cle.minimum_images(test1.astype(np.uint16), test2.astype(np.float32))
    result5 = cle.minimum_images(test1.astype(np.uint32), test2.astype(np.float32))

    assert np.array_equal(result1, reference)
    assert np.array_equal(result2, reference)
    assert np.array_equal(result3, reference)
    assert np.array_equal(result4, reference)
    assert np.array_equal(result5, reference)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_nan_to_num():
    data = np.asarray([[-np.inf], [np.inf], [np.nan], [0], [1]])

    print(np.nan_to_num(data, nan=3, posinf=4, neginf=5))
    print(cle.nan_to_num(data, nan=3, posinf=4, neginf=5))

    assert np.allclose(
        np.nan_to_num(data, nan=3, posinf=4, neginf=5),
        cle.nan_to_num(data, nan=3, posinf=4, neginf=5),
    )


def test_nan_to_num_defaults():
    data = np.asarray([[-np.inf], [np.inf], [np.nan], [0], [1]])

    print(np.nan_to_num(data))
    print(cle.nan_to_num(data, nan=0, posinf=np.inf, neginf=-np.inf))

    result = cle.nan_to_num(data, nan=0, posinf=np.inf, neginf=-np.inf)
    print(result)

    assert result[0, 0] < 0  # a very small number
    assert result[1, 0] > 0  # a very large number
    assert result[2, 0] == 0
    assert result[3, 0] == 0
    assert result[4, 0] == 1



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_detect_maxima_box():
    gpu_input = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [1, 0, 0, 2, 0],
                [1, 2, 1, 3, 2],
                [0, 1, 0, 2, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )
    gpu_output = cle.create_like(gpu_input)

    gpu_reference = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [0, 1, 0, 1, 0],
                [0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    cle.detect_maxima(gpu_input, gpu_output)

    a = cle.pull(gpu_output)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_array_equal():
    input1 = np.asarray([1, 2, 3])
    input2 = np.asarray([4, 5, 7])
    input3 = np.asarray([1, 2, 3, 3])
    input4 = np.asarray([1.0, 2.0, 3.0])

    assert not cle.array_equal(input1, input2)
    assert not cle.array_equal(input1, input3)
    assert cle.array_equal(input1, input4)



from os import path
from pathlib import Path
from typing import Optional, Union, Callable
from inspect import getmembers, isfunction

import numpy as np
from matplotlib.colors import ListedColormap

from ._array import Array, Image
from ._core import Device, get_device
from ._memory import pull
from ._pyclesperanto import _execute, _native_execute


def execute(
    anchor=__file__,
    kernel_source: str = "",
    kernel_name: str = "",
    global_size: tuple = (1, 1, 1),
    parameters: dict = {},
    constants: dict = {},
    device: Device = None,
):
    """Execute a kernel from a file or a string

    Call, build, and execute a kernel compatible with CLIj framework.
    The kernel can be called from a file or a string.

    Parameters
    ----------
    anchor : str, default = '__file__'
        Enter __file__ when calling this method and the corresponding open.cl
        file lies in the same folder as the python file calling it.
        Ignored if kernel_source is a string.
    kernel_source : str
        Filename of the open.cl file to be called or string containing the open.cl source code
    kernel_name : str
        Kernel method inside the open.cl file to be called
        most clij/clesperanto kernel functions have the same name as the file they are in
    global_size : tuple (z,y,x), default = (1, 1, 1)
        Global_size according to OpenCL definition (usually shape of the destination image).
    parameters : dict(str, [Array, float, int])
        Dictionary containing parameters. Take care: They must be of the
        right type and in the right order as specified in the open.cl file.
    constants: dict(str, int), optional
        Dictionary with names/values which will be added to the define
        statements. They are necessary, e.g. to create arrays of a given
        maximum size in OpenCL as variable array lengths are not supported.
    device : Device, default = None
        The device to execute the kernel on. If None, use the current device
    """

    # load the kernel file
    def load_file(anchor, filename):
        """Load the opencl kernel file as a string"""
        if anchor is None:
            kernel = Path(filename).read_text()
        else:
            kernel = (Path(anchor).parent / filename).read_text()
        return kernel

    # test if kernel_source ends with .cl or .cu
    if kernel_source.endswith(".cl") or kernel_source.endswith(".cu"):
        kernel_source = load_file(anchor, kernel_source)

    # manage the device if not given
    if not device:
        device = get_device()

    # manage global range
    if not isinstance(global_size, tuple):
        if isinstance(global_size, list) or isinstance(global_size, np.ndarray):
            global_size = tuple(global_size)
        else:
            global_size = (global_size,)

    _execute(device, kernel_name, kernel_source, parameters, global_size, constants)


def native_execute(
    anchor=__file__,
    kernel_source: str = "",
    kernel_name: str = "",
    global_size: tuple = (1, 1, 1),
    local_size: tuple = (1, 1, 1),
    parameters: dict = {},
    device: Device = None,
):
    """Execute an OpenCL kernel from a file or a string

    Call, build, and execute a kernel compatible with OpenCL language.
    The kernel can be called from a file or a string.

    The parameters must still be passed as a dictionary with the correct types and order.
    Buffer parameters must be passed as Array objects. Scalars must be passed as Python native float or int.

    Warning: Only 1D buffers are supported for now.

    Parameters
    ----------
    anchor : str, default = '__file__'
        Enter __file__ when calling this method and the corresponding open.cl
        file lies in the same folder as the python file calling it.
        Ignored if kernel_source is a string.
    kernel_source : str
        Filename of the open.cl file to be called or string containing the open.cl source code
    kernel_name : str
        Kernel method inside the open.cl file to be called
        most clij/clesperanto kernel functions have the same name as the file they are in
    global_size : tuple (z,y,x), default = (1, 1, 1)
        Global_size according to OpenCL definition (usually shape of the destination image).
    local_size : tuple (z,y,x), default = (1, 1, 1)
        Local_size according to OpenCL definition (usually default is good).
    parameters : dict(str, [Array, float, int])
        Dictionary containing parameters. Take care: They must be of the
        right type and in the right order as specified in the open.cl file.
    device : Device, default = None
        The device to execute the kernel on. If None, use the current device
    """

    # load the kernel file
    def load_file(anchor, filename):
        """Load the opencl kernel file as a string"""
        if anchor is None:
            kernel = Path(filename).read_text()
        else:
            kernel = (Path(anchor).parent / filename).read_text()
        return kernel

    # test if kernel_source ends with .cl or .cu
    if kernel_source.endswith(".cl") or kernel_source.endswith(".cu"):
        kernel_source = load_file(anchor, kernel_source)

    # manage the device if not given
    if not device:
        device = get_device()

    # manage global range
    if not isinstance(global_size, tuple):
        if isinstance(global_size, list) or isinstance(global_size, np.ndarray):
            global_size = tuple(global_size)
        else:
            global_size = (global_size,)

    # manage local range
    if not isinstance(local_size, tuple):
        if isinstance(local_size, list) or isinstance(local_size, np.ndarray):
            local_size = tuple(local_size)
        else:
            local_size = (local_size,)

    _native_execute(
        device, kernel_name, kernel_source, parameters, global_size, local_size
    )


def imshow(
    image,
    title: Optional[str] = None,
    labels: Optional[bool] = False,
    min_display_intensity: Optional[float] = None,
    max_display_intensity: Optional[float] = None,
    color_map: Optional[str] = None,
    plot=None,
    colorbar: Optional[bool] = False,
    colormap: Union[str, ListedColormap, None] = None,
    alpha: Optional[float] = None,
    continue_drawing: Optional[bool] = False,
):
    """Visualize an image, e.g. in Jupyter notebooks using matplotlib.

    Parameters
    ----------
    image: np.ndarray
        numpy or OpenCL-backed image to visualize
    title: str, optional
        Obsolete (kept for ImageJ-compatibility)
    labels: bool, optional
        True: integer labels will be visualized with colors
        False: Specified or default colormap will be used to display intensities.
    min_display_intensity: float, optional
        lower limit for display range
    max_display_intensity: float, optional
        upper limit for display range
    color_map: str, optional
        deprecated, use colormap instead
    plot: matplotlib axis, optional
        Plot object where the image should be shown. Useful for putting multiple images in subfigures.
    colorbar: bool, optional
        True puts a colorbar next to the image. Will not work with label images and when visualizing multiple
        images (continue_drawing=True).
    colormap: str or matplotlib colormap, optional
    alpha: float, optional
        alpha blending value
    continue_drawing: float
        True: the next shown image can be visualized on top of the current one, e.g. with alpha = 0.5
    """
    if len(image.shape) == 3:
        from ._tier1 import maximum_z_projection

        image = pull(maximum_z_projection(image))

    image = pull(image)

    if color_map is not None:
        import warnings

        warnings.warn(
            "The imshow parameter color_map is deprecated. Use colormap instead."
        )
        if colormap is None:
            colormap = color_map

    if labels:
        if not hasattr(imshow, "colormap"):
            from numpy.random import MT19937, RandomState, SeedSequence

            rs = RandomState(MT19937(SeedSequence(3)))
            lut = rs.rand(65537, 3)
            lut[0, :] = 0
            # these are the first four colours from matplotlib's default
            lut[1] = [0.12156862745098039, 0.4666666666666667, 0.7058823529411765]
            lut[2] = [1.0, 0.4980392156862745, 0.054901960784313725]
            lut[3] = [0.17254901960784313, 0.6274509803921569, 0.17254901960784313]
            lut[4] = [0.8392156862745098, 0.15294117647058825, 0.1568627450980392]
            colormap = ListedColormap(lut)

        if min_display_intensity is None:
            min_display_intensity = 0
        if max_display_intensity is None:
            max_display_intensity = 65536

    if colormap is None:
        colormap = "Greys_r"

    cmap = colormap
    if plot is None:
        import matplotlib.pyplot as plt

        plt.imshow(
            image,
            cmap=cmap,
            vmin=min_display_intensity,
            vmax=max_display_intensity,
            interpolation="nearest",
            alpha=alpha,
        )
        if colorbar:
            plt.colorbar()
        if not continue_drawing:
            plt.show()
    else:
        plot.imshow(
            image,
            cmap=cmap,
            vmin=min_display_intensity,
            vmax=max_display_intensity,
            interpolation="nearest",
            alpha=alpha,
        )
        if colorbar:
            plot.colorbar()
    if title is not None:
        plt.title(title)


def operations(
    must_have_categories: list = None, must_not_have_categories: list = None
) -> dict:
    """Retrieve a dictionary of operations, which can be filtered by annotated categories.

    Parameters
    ----------
    must_have_categories : list of str, optional
        if provided, the result will be filtered so that operations must contain all given categories.
    must_not_have_categories : list of str, optional
        if provided, the result will be filtered so that operations must not contain all given categories.

    Returns
    -------
    dict of str : Callable function
    """

    import pyclesperanto as cle

    if isinstance(must_have_categories, str):
        must_have_categories = [must_have_categories]
    if isinstance(must_not_have_categories, str):
        must_have_categories = [must_not_have_categories]

    result = {}

    # retrieve all operations and cache the result for later reuse
    if not hasattr(operations, "_all") or operations._all is None:
        operations._all = getmembers(cle, isfunction)

    # filter operations according to given constraints
    for operation_name, operation in operations._all:
        keep_it = True
        if hasattr(operation, "categories") and operation.categories is not None:
            if must_have_categories is not None:
                if not all(
                    item in operation.categories for item in must_have_categories
                ):
                    keep_it = False

            if must_not_have_categories is not None:
                if any(
                    item in operation.categories for item in must_not_have_categories
                ):
                    keep_it = False
        else:
            if must_have_categories is not None:
                keep_it = False
        if keep_it:
            result[operation_name] = operation

    return result


def operation(name: str) -> Callable:
    """Returns a function from the pyclesperanto package

    Parameters
    ----------
    name : str
        name of the operation

    Returns
    -------
        Callable function
    """
    dict = operations()
    return dict[name]


def search_operation_names(name: str) -> list:
    """
    Returns a list of operation names containing the given string

    Parameters
    ----------
    name : str
        string to search for in operation names

    Returns
    -------
    list
        list of operation names containing the given string
    """
    return [a for a in list(operations().keys()) if name in a]



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_set_ramp_y():
    result = cle.push(
        np.asarray(
            [[[0, 0, 0], [3, 4, 3], [3, 4, 3]], [[3, 4, 3], [3, 4, 3], [3, 4, 3]]]
        )
    )

    reference = cle.push(
        np.asarray(
            [[[0, 0, 0], [1, 1, 1], [2, 2, 2]], [[0, 0, 0], [1, 1, 1], [2, 2, 2]]]
        )
    )

    cle.set_ramp_y(result)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.allclose(a, b, 0.001)



import numpy as np
import pytest

import pyclesperanto as cle

cle.select_device("TX")


def test_dask_compatibility():
    array = pytest.importorskip("dask.array")

    np_arr = np.random.random((100, 100))
    da_arr = array.from_array(np_arr)
    cl_arr = cle.asarray(np_arr)

    a = cl_arr + np_arr
    b = cle.add_images(cl_arr, da_arr)
    c = np_arr + np_arr

    assert np.allclose(c, a)
    assert np.allclose(c, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def onlyzero_overwrite_maximum_sphere():
    test1 = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0],
                [0, 1, 2, 3, 0],
                [0, 2, 3, 4, 0],
                [0, 4, 4, 5, 0],
                [0, 0, 0, 0, 0],
            ]
        )
    )

    reference = cle.push(
        np.asarray(
            [
                [0, 1, 2, 3, 0],
                [1, 1, 2, 3, 3],
                [2, 2, 3, 4, 4],
                [4, 4, 4, 5, 5],
                [0, 4, 4, 5, 0],
            ]
        )
    )

    result = cle.create(test1)
    flag = cle.create((1, 1, 1))
    cle.onlyzero_overwrite_maximum(test1, flag, result, connectivity="sphere")

    print(result)

    a = cle.pull(result)
    b = cle.pull(reference)
    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_voronoi_otsu_labeling():
    gpu_input = cle.push(
        np.asarray(
            [
                [0, 0, 1, 1, 0, 0],
                [0, 1, 8, 9, 1, 0],
                [0, 1, 7, 6, 1, 0],
                [0, 0, 1, 1, 1, 0],
                [0, 0, 1, 1, 1, 0],
                [0, 0, 1, 8, 7, 1],
                [0, 0, 1, 1, 1, 0],
            ]
        )
    )

    gpu_reference = cle.push(
        np.asarray(
            [
                [0, 0, 1, 1, 0, 0],
                [0, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 0],
                [0, 0, 1, 2, 2, 0],
                [0, 0, 0, 2, 2, 0],
                [0, 0, 2, 2, 2, 0],
                [0, 0, 0, 2, 2, 0],
            ]
        )
    )

    gpu_output = cle.voronoi_otsu_labeling(gpu_input, spot_sigma=1, outline_sigma=1)

    a = cle.pull(gpu_output)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)



import pyclesperanto as cle
import numpy as np


def test_exclude_labels_out_of_size_range_2d():

    gpu_input = cle.push(
        np.asarray(
            [
                [1, 1, 2, 0, 3, 3],
                [1, 1, 2, 0, 3, 3],
                [0, 0, 0, 0, 0, 0],
                [4, 4, 5, 6, 6, 6],
                [4, 4, 5, 6, 6, 6],
            ]
        )
    )

    gpu_reference = cle.push(
        np.asarray(
            [
                [1, 1, 0, 0, 2, 2],
                [1, 1, 0, 0, 2, 2],
                [0, 0, 0, 0, 0, 0],
                [3, 3, 0, 0, 0, 0],
                [3, 3, 0, 0, 0, 0],
            ]
        )
    )

    gpu_output = cle.filter_label_by_size(gpu_input, min_size=4, max_size=5)

    a = cle.pull(gpu_output)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)


def test_exclude_small_labels_2d():
    gpu_input = cle.push(
        np.asarray(
            [
                [1, 1, 2, 0, 3, 3],
                [1, 1, 2, 0, 3, 3],
                [0, 0, 0, 0, 0, 0],
                [4, 4, 5, 6, 6, 6],
                [4, 4, 5, 6, 6, 6],
            ]
        )
    )

    gpu_reference = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0],
                [0, 0, 0, 0, 0, 0],
                [0, 0, 0, 1, 1, 1],
                [0, 0, 0, 1, 1, 1],
            ]
        )
    )

    gpu_output = cle.exclude_small_labels(gpu_input, max_size=5)

    a = cle.pull(gpu_output)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)


def test_exclude_large_labels_2d():
    gpu_input = cle.push(
        np.asarray(
            [
                [1, 1, 2, 0, 3, 3],
                [1, 1, 2, 0, 3, 3],
                [0, 0, 0, 0, 0, 0],
                [4, 4, 5, 6, 6, 6],
                [4, 4, 5, 6, 6, 6],
            ]
        )
    )

    gpu_reference = cle.push(
        np.asarray(
            [
                [0, 0, 1, 0, 0, 0],
                [0, 0, 1, 0, 0, 0],
                [0, 0, 0, 0, 0, 0],
                [0, 0, 2, 0, 0, 0],
                [0, 0, 2, 0, 0, 0],
            ]
        )
    )

    gpu_output = cle.exclude_large_labels(gpu_input, min_size=3)

    a = cle.pull(gpu_output)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_transpose_yz():
    test1 = cle.push(np.asarray([[[0, 1], [2, 3]], [[4, 5], [6, 7]]]))

    reference = cle.push(np.asarray([[[0, 1], [4, 5]], [[2, 3], [6, 7]]]))

    result = cle.create(test1)
    cle.transpose_yz(test1, result)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)


def test_transpose_yz_3d_generate_output():
    test1 = cle.push(np.asarray([[[1, 2, 6], [3, 4, 5]]]))

    result = cle.transpose_yz(test1)

    a = cle.pull(result)

    assert np.min(a) == 1
    assert np.max(a) == 6
    assert np.mean(a) == 3.5


def test_transpose_yz_2d_generate_output():
    test1 = cle.push(np.asarray([[1, 2, 6], [3, 4, 5]]))

    result = cle.transpose_yz(test1)

    a = cle.pull(result)

    print(a)

    assert np.min(a) == 1
    assert np.max(a) == 6
    assert np.mean(a) == 3.5


def test_transpose_yz_1d_generate_output():
    test1 = cle.push(np.asarray([1, 2, 6, 3, 4, 5]))

    result = cle.transpose_yz(test1)

    a = cle.pull(result)

    print(a)

    assert np.min(a) == 1
    assert np.max(a) == 6
    assert np.mean(a) == 3.5



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")

arr1 = cle.push(np.asarray([1, 2, 3, 4, 5]))
arr2 = cle.push(np.asarray([[1, 2, 3, 4, 5], [6, 7, 8, 9, 10]]))
arr3 = cle.push(
    np.asarray(
        [
            [[1, 2, 3, 4, 5], [6, 7, 8, 9, 10]],
            [[11, 12, 13, 14, 15], [16, 17, 18, 19, 20]],
        ]
    )
)


def test_max():
    assert arr1.max() == 5
    assert arr2.max() == 10
    assert arr3.max() == 20


def test_min():
    assert arr1.min() == 1
    assert arr2.min() == 1
    assert arr3.min() == 1


def test_sum():
    assert arr1.sum() == 15
    assert arr2.sum() == 55
    assert arr3.sum() == 210



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_flag_existing_labels():
    gpu_input = cle.push(np.asarray([[1, 2, 3], [1, 6, 6], [7, 8, 9]]))

    gpu_reference = cle.push(np.asarray([0, 1, 1, 1, 0, 0, 1, 1, 1, 1]))

    gpu_output = cle.flag_existing_labels(gpu_input)

    a = cle.pull(gpu_output)
    b = cle.pull(gpu_reference)
    print(gpu_input)
    print(a)
    print(b)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")

source = np.asarray([0, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0])
reference_4 = np.asarray([2, 2, 1])
reference_2 = np.asarray([1, 1, 0, 2, 0, 1])


def sum_reduction(source, blocksize):
    flagged_indices = cle.push(source)

    block_sums = cle.sum_reduction_x(flagged_indices, blocksize=blocksize)

    return cle.pull(block_sums)


def test_sum_reduction():
    result = sum_reduction(source, 4)
    print(result)
    print(reference_4)
    assert np.array_equal(result, reference_4)

    result = sum_reduction(source, 2)
    print(result)
    print(reference_2)
    assert np.array_equal(result, reference_2)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_set_nonzero_pixels_to_pixelindex():
    test1 = cle.push(
        np.asarray([[0, 0, 0, 1], [0, 0, 3, 1], [0, 0, 3, 1], [1, 1, 1, 1]])
    )

    reference = cle.push(
        np.asarray([[0, 0, 0, 3], [0, 0, 6, 7], [0, 0, 10, 11], [12, 13, 14, 15]])
    )

    result = cle.create(test1)
    cle.set_nonzero_pixels_to_pixelindex(test1, result, offset=0)

    a = cle.pull(result)
    b = cle.pull(reference)
    print(a)
    print(b)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_push_np():
    reference = np.asarray([[1, 2], [-3, 4]])

    image = cle.push(reference)

    result = cle.pull(image)

    assert np.allclose(result, reference)


def test_push_list():
    reference = [[1, 2], [-3, 4]]

    image = cle.push(reference)

    result = cle.pull(image)

    assert np.allclose(result, reference)


def test_push_tuple():
    reference = ([1, 2], [-3, 4])

    image = cle.push(reference)

    result = cle.pull(image)

    assert np.allclose(result, reference)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_iterator():
    image = np.random.random((3, 2, 5))
    cle_image = cle.asarray(image)
    print(image)
    print(cle_image)

    assert cle.array_equal(image, cle_image)

    for i, j in zip(image, cle_image):
        print(cle.array_equal(i, j))
        assert cle.array_equal(i, j)


def test_enumerate():
    cle_array = cle.create((2, 10))
    cle.set_ramp_x(cle_array)

    sum_ = 0
    for i, y in enumerate(cle_array[0]):
        print(i, y)
        sum_ += y

    assert sum_ == 45


def test_zip():
    cle_array = cle.create((2, 10))
    cle.set_ramp_x(cle_array)

    sum_ = 0
    for x, y in zip(cle_array[0], cle_array[1]):
        print(x, y)
        sum_ += y

    assert sum_ == 45


# def test_iter_centroids():
#     labels = cle.asarray(
#         [
#             [0, 0, 0, 0, 0],
#             [0, 1, 0, 3, 0],
#             [0, 0, 0, 0, 0],
#             [0, 0, 2, 0, 0],
#             [0, 0, 0, 0, 0],
#         ]
#     )

#     centroids = cle.centroids_of_labels(labels)

#     print(centroids)

#     for i, j in zip(centroids[0], centroids[1]):
#         print(i, j)



from inspect import getmembers, isfunction


def categories() -> list:
    """
    Returns a list of all categories of operations in pyclesperanto

    Returns
    -------
    list
        A list of strings with category names
    """

    import pyclesperanto as cle

    ops = getmembers(cle, isfunction)
    cats = {
        cat
        for _, operation in ops
        if hasattr(operation, "categories") and operation.categories is not None
        for cat in operation.categories
    }
    return list(cats)



import pyclesperanto as cle
import numpy as np


def test_label_pixel_count_map():

    labels = cle.push(np.asarray([[1, 1, 2], [1, 0, 0], [3, 3, 0]]))

    reference = cle.push(np.asarray([[3, 3, 1], [3, 0, 0], [2, 2, 0]]))

    result = cle.label_pixel_count_map(labels)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)


def test_label_pixel_count_map_3d():

    labels = cle.push(
        np.asarray(
            [
                [
                    [1, 1, 2],
                ],
                [
                    [1, 0, 0],
                ],
                [
                    [3, 3, 0],
                ],
            ]
        )
    )

    reference = cle.push(
        np.asarray(
            [
                [
                    [3, 3, 1],
                ],
                [
                    [3, 0, 0],
                ],
                [
                    [2, 2, 0],
                ],
            ]
        )
    )

    result = cle.label_pixel_count_map(labels)

    a = cle.pull(result)
    b = cle.pull(reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_relabel_sequential():
    gpu_input = cle.push(np.asarray([[1, 2, 3], [1, 6, 6], [7, 8, 9]]))
    gpu_output = cle.create_like(gpu_input)

    gpu_reference = cle.push(np.asarray([[1, 2, 3], [1, 4, 4], [5, 6, 7]]))

    cle.relabel_sequential(gpu_input, gpu_output, 2048)

    a = cle.pull(gpu_output)
    b = cle.pull(gpu_reference)

    print(a)
    print(b)

    assert np.array_equal(a, b)



import numpy as np
from skimage.segmentation.morphsnakes import inf_sup, sup_inf

import pyclesperanto as cle

cle.select_device("TX")


def test_superior_inferior_2d():
    test = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 0, 0, 0, 0, 1, 1, 0],
                [0, 1, 1, 0, 0, 0, 0, 1, 1, 0],
                [0, 1, 1, 0, 0, 0, 0, 1, 1, 0],
                [0, 1, 1, 0, 0, 0, 0, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ]
        ).astype(np.uint8)
    )

    reference = cle.push(
        np.asarray(
            [
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                [0, 0, 1, 1, 1, 1, 1, 1, 0, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 1, 1, 0, 0, 0, 0, 1, 1, 0],
                [0, 1, 1, 0, 0, 0, 0, 1, 1, 0],
                [0, 1, 1, 0, 0, 0, 0, 1, 1, 0],
                [0, 1, 1, 0, 0, 0, 0, 1, 1, 0],
                [0, 1, 1, 1, 1, 1, 1, 1, 1, 0],
                [0, 0, 1, 1, 1, 1, 1, 1, 0, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            ]
        ).astype(np.uint8)
    )

    result = cle.create(test)
    cle.binary_supinf(test, result)

    print(result)

    a = cle.pull(result)
    b = cle.pull(reference)
    assert np.array_equal(a, b)

    reference2 = sup_inf(test)
    assert cle.array_equal(result, reference2)


def test_superior_inferior_2d_compare_with_skimage_x():

    array = np.zeros((100, 85), dtype=np.uint8)
    array[15:85, 15:85] = 1
    array[35:65, 35:65] = 0

    result = cle.binary_supinf(cle.binary_infsup(array))
    reference = sup_inf(inf_sup(array))

    print("result", result)
    print("reference", reference)

    assert cle.array_equal(result, reference)


def test_superior_inferior_2d_compare_with_skimage_y():

    array = np.zeros((85, 100), dtype=np.uint8)
    array[15:85, 15:85] = 1
    array[35:65, 35:65] = 0

    result = cle.binary_supinf(cle.binary_infsup(array))
    reference = sup_inf(inf_sup(array))

    print("result", result)
    print("reference", reference)

    assert cle.array_equal(result, reference)


def test_superior_inferior_3d():
    test = cle.push(
        np.asarray(
            [
                [
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                ],
                [
                    [0, 0, 0, 0, 0],
                    [0, 1, 1, 1, 0],
                    [0, 1, 1, 1, 0],
                    [0, 1, 1, 1, 0],
                    [0, 0, 0, 0, 0],
                ],
                [
                    [0, 0, 0, 0, 0],
                    [0, 1, 1, 1, 0],
                    [0, 1, 0, 1, 0],
                    [0, 1, 1, 1, 0],
                    [0, 0, 0, 0, 0],
                ],
                [
                    [0, 0, 0, 0, 0],
                    [0, 1, 1, 1, 0],
                    [0, 1, 1, 1, 0],
                    [0, 1, 1, 1, 0],
                    [0, 0, 0, 0, 0],
                ],
                [
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0],
                ],
            ]
        ).astype(np.uint8)
    )

    reference = inf_sup(test)

    result = cle.binary_infsup(test)

    print("result", result)
    print("reference", reference)

    a = cle.pull(result)
    b = cle.pull(reference)
    assert np.array_equal(a, b)


def test_superior_inferior_3d_compare_with_skimage_x():

    array = np.zeros((5, 5, 4), dtype=np.uint8)
    array[1:4, 1:4, 1:4] = 1
    array[2, 2, 2] = 0

    result = cle.binary_supinf(cle.binary_infsup(array))
    reference = sup_inf(inf_sup(array))

    print("result", result)
    print("reference", reference)

    assert cle.array_equal(result, reference)


def test_superior_inferior_3d_compare_with_skimage_y():

    array = np.zeros((5, 4, 5), dtype=np.uint8)
    array[1:4, 1:4, 1:4] = 1
    array[2, 2, 2] = 0

    result = cle.binary_supinf(cle.binary_infsup(array))
    reference = sup_inf(inf_sup(array))

    print("result", result)
    print("reference", reference)

    assert cle.array_equal(result, reference)


def test_superior_inferior_3d_compare_with_skimage_z():

    array = np.zeros((4, 5, 5), dtype=np.uint8)
    array[1:4, 1:4, 1:4] = 1
    array[2, 2, 2] = 0

    result = cle.binary_supinf(cle.binary_infsup(array))
    reference = sup_inf(inf_sup(array))

    print("result", result)
    print("reference", reference)

    assert cle.array_equal(result, reference)



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_minimum_of_masked_pixels_mini_x():
    np_input = np.asarray([1, 2, 3, 4])
    np_mask = np.asarray([0, 1, 1, 0])

    gpu_input = cle.push(np_input)
    gpu_mask = cle.push(np_mask)

    result = cle.minimum_of_masked_pixels(gpu_input, gpu_mask)
    print(result)
    assert result == 2


def test_minimum_of_masked_pixels_mini_y():
    np_input = np.asarray([[1], [2], [3], [4]])
    np_mask = np.asarray([[0], [1], [1], [0]])

    gpu_input = cle.push(np_input)
    gpu_mask = cle.push(np_mask)

    result = cle.minimum_of_masked_pixels(gpu_input, gpu_mask)
    print(result)
    assert result == 2


def test_minimum_of_masked_pixels():
    np_input = np.asarray(
        [
            [[1, 2, 3, 10], [4, 5, 6, 11], [7, 8, 9, 12]],
            [[1, 2, 3, 13], [4, 5, 6, 14], [7, 8, 9, 15]],
        ]
    )

    np_mask = np.asarray(
        [
            [[0, 0, 0, 0], [0, 1, 1, 0], [0, 1, 1, 0]],
            [[0, 1, 1, 0], [0, 1, 1, 0], [0, 0, 0, 0]],
        ]
    )

    gpu_input = cle.push(np_input)
    gpu_mask = cle.push(np_mask)

    print("gpu_input")
    print(gpu_input)
    print("gpu_mask")
    print(gpu_mask)

    result = cle.minimum_of_masked_pixels(gpu_input, gpu_mask)
    print(result)
    assert result == 2

    gpu_input = cle.push(np_input)
    gpu_mask = cle.push(np_mask)

    result = cle.minimum_of_masked_pixels(gpu_input, gpu_mask)
    print(result)
    assert result == 2



import numpy as np

import pyclesperanto as cle

cle.select_device("TX")


def test_large_hessian_eigenvalue_2d():
    test = np.asarray([[1, -1], [1, -1]])

    reference_large_hessian_eigenvalue = np.asarray([[0, 2], [0, 2]])

    large_hessian_eigenvalue = cle.large_hessian_eigenvalue(test)

    print(large_hessian_eigenvalue)

    assert np.allclose(reference_large_hessian_eigenvalue, large_hessian_eigenvalue)


def test_large_hessian_eigenvalue_3d():
    test = np.asarray(
        [
            [[1, -1], [1, -1]],
            [[2, -2], [2, -2]],
        ]
    )

    reference_large_hessian_eigenvalue = np.asarray(
        [
            [[1.1, 2.1], [1.1, 2.1]],
            [[0, 4.1], [0, 4.1]],
        ]
    )

    large_hessian_eigenvalue = cle.large_hessian_eigenvalue(test)

    print(large_hessian_eigenvalue)

    assert np.allclose(
        reference_large_hessian_eigenvalue, large_hessian_eigenvalue, atol=0.1
    )
